dicom_tags = {
    "accessionnumber": "0008|0050",
    "acquisitiondate": "0008|0022",
    "acquisitionmatrix": "0018|1310",
    "acquisitiontime": "0008|0032",
    "additionalpatienthistory": "0010|21b0",
    "bitsallocated": "0028|0100",
    "bitsstored": "0028|0101",
    "bodypartexamined": "0018|0015",
    "columns": "0028|0011",
    "contentdate": "0008|0023",
    "contenttime": "0008|0033",
    "deidentificationmethod": "0012|0063",
    "deviceserialnumber": "0018|1000",
    "diffusionbvalue": "0018|9087",
    "echotime": "0018|0081",
    "echotrainlength": "0018|0091",
    "flipangle": "0018|1314",
    "frameofreferenceuid": "0020|0052",
    "imagednucleus": "0018|0085",
    "imageorientationpatient": "0020|0037",
    "imagepositionpatient": "0020|0032",
    "imagetype": "0008|0008",
    "imagingfrequency": "0018|0084",
    "inplanephaseencodingdirection": "0018|1312",
    "instancecreationdate": "0008|0012",
    "instancecreationtime": "0008|0013",
    "institutionaldepartmentname": "0008|1040",
    "largestimagepixelvalue": "0028|0107",
    "magneticfieldstrength": "0018|0087",
    "manufacturer": "0008|0070",
    "manufacturersmodelname": "0008|1090",
    "modality": "0008|0060",
    "mracquisitiontype": "0018|0023",
    "numberofaverages": "0018|0083",
    "numberofphaseencodingsteps": "0018|0089",
    "patientid": "0010|0020",
    "patientidentityremoved": "0012|0062",
    "patientposition": "0018|5100",
    "patientsage": "0010|1010",
    "patientsbirthdate": "0010|0030",
    "patientssex": "0010|0040",
    "percentphasefieldofview": "0018|0094",
    "percentsampling": "0018|0093",
    "performedprocedurestepdescription": "0040|0254",
    "performedprocedurestepstartdate": "0040|0244",
    "pixelspacing": "0028|0030",
    "protocolname": "0018|1030",
    "repetitiontime": "0018|0080",
    "rescaleintercept": "0028|1052",
    "rescaleslope": "0028|1053",
    "rows": "0028|0010",
    "samplesperpixel": "0028|0002",
    "scanningsequence": "0018|0020",
    "scanoptions": "0018|0022",
    "sequencename": "0018|0024",
    "sequencevariant": "0018|0021",
    "seriesdate": "0008|0021",
    "seriesdescription": "0008|103e",
    "seriesinstanceuid": "0020|000e",
    "seriestime": "0008|0031",
    "slicelocation": "0020|1041",
    "slicethickness": "0018|0050",
    "smallestimagepixelvalue": "0028|0106",
    "sopclassuid": "0008|0016",
    "sopinstanceuid": "0008|0018",
    "spacingbetweenslices": "0018|0088",
    "specificcharacterset": "0008|0005",
    "stationname": "0008|1010",
    "studydate": "0008|0020",
    "studydescription": "0008|1030",
    "studyid": "0020|0010",
    "studyinstanceuid": "0020|000d",
    "studytime": "0008|0030",
    "variableflipangleflag": "0018|1315",
    "windowcenter": "0028|1050",
    "windowwidth": "0028|1051"
}
